import numpy as np
from scipy.io import loadmat
import matplotlib.pyplot as plt

from matplotlib import pylab as plt
from scipy.special import ellipj, ellipk

import h5py
import torch
from torch.utils.data import TensorDataset, DataLoader
import sys

class Logger(object):
    def __init__(self, fileN="default.log"):
        self.terminal = sys.stdout
        self.log = open(fileN, "a")
 
    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)
 
    def flush(self):
        pass

#******************************************************************************
# Read in data
#******************************************************************************

def data_from_name(name, noise = 0.0, theta=2.4, p = '4581'):
    if name == 'pendulum_lin':
        return pendulum_lin(noise)      
    if name == 'pendulum':
        return pendulum(noise, theta)
    if name == 'pendulum_timeseries':
        return pendulum_timeseries(noise, theta)
    if name == 'lorenz':
        return lorenz()
    if name == 'flow':
        return flow_cylinder()
    if name == 'shear_stress':
        return shear_stress(p)
    else:
        raise ValueError('dataset {} not recognized'.format(name))

def lorenz(h = 0.001, n = 5000):
    xs, ys, zs = [], [], []

    a, b, c = 10.0, 28.0, 8.0 / 3.0                                    
    x0, y0, z0 = 0.1, 0, 0
    for i in range(n):
        x1 = x0 + h * a * (y0 - x0)
        y1 = y0 + h * (x0 * (b - z0) - y0)
        z1 = z0 + h * (x0 * y0 - c * z0)
        x0, y0, z0 = x1, y1, z1
        xs.append(x0)
        ys.append(y0)
        zs.append(z0)
    X=[]
    X.append(xs)
    X.append(ys)
    X.append(zs)
    X=np.array(X).T
    # print(X.shape)

    # Q = np.random.standard_normal((96,3))
    # Q,_ = np.linalg.qr(Q)

    # X = X.dot(Q.T) # rotate


    # X = 2 * (X - np.min(X)) / np.ptp(X) - 1
    # train_num = 1000

    # X_train = X[:train_num]   
    # X_test = X[train_num:]

    # print(X_train.shape)
    return X

def rescale(Xsmall, Xsmall_test):
    #******************************************************************************
    # Rescale data
    #******************************************************************************
    Xmin = Xsmall.min()
    Xmax = Xsmall.max()
    
    Xsmall = ((Xsmall - Xmin) / (Xmax - Xmin)) 
    Xsmall_test = ((Xsmall_test - Xmin) / (Xmax - Xmin)) 

    return Xsmall, Xsmall_test

def timedelay(x, d, tau = 1):
    '''
    x: time series
    d: target dimension after time delay
    m: embedded dimension, also time window
    tau: delay time
    '''
    N = max(x.shape)
    m = int(np.floor((N-d)/tau) + 1) # tau = 1 usually

    if 2*d >= N:
        print('Target dimension exceed')
        return 0

    H = np.zeros((d, N-d+1))
    for i in range(d):
        H[i, :] = x[i: i+ N-d + 1: tau]
    
    return H, m

def embed_1dim(x,tau=1,m=2):
	# Check inputs shape
	if len(x.shape)!=1:
		if len(x.shape)==2:
			Nx, Ny = x.shape
			if Nx==1 and Ny==1:
				raise ValueError("The input x must be a time series with more than one value.")
			else:
				if Nx==1:
					x = x[0,:]
				else:
					x = x[:,0]
		else:
			raise ValueError("The input x must be 1-dimensional.")
	if not isinstance(tau, int):
		raise ValueError("The input tau must be a scalar integer.")
	if not isinstance(m, int):
		raise ValueError("The input m must be a scalar integer.")
	# Now that x is 1-dim, extract number of epochs
	Nt = x.shape[0]
	# Create Hankel matrix
	H = np.zeros((m,Nt-(m-1)*tau))
	for mm in range(m):
		if mm==m-1:
			H[mm,:] = x[mm*tau:].T
		else:
			H[mm,:] = x[mm*tau:-((m-mm-1)*tau)].T
	return H

def pendulum(noise, theta=2.4):
    
    np.random.seed(1)

    def sol(t,theta0):
        S = np.sin(0.5*(theta0) )
        K_S = ellipk(S**2)
        omega_0 = np.sqrt(9.81)
        sn,cn,dn,ph = ellipj( K_S - omega_0*t, S**2 )
        theta = 2.0*np.arcsin( S*sn )
        d_sn_du = cn*dn
        d_sn_dt = -omega_0 * d_sn_du
        d_theta_dt = 2.0*S*d_sn_dt / np.sqrt(1.0-(S*sn)**2)
        return np.stack([theta, d_theta_dt],axis=1)
    
    
    anal_ts = np.arange(0, 2200*0.1, 0.1)
    X = sol(anal_ts, theta)
    
    X = X.T
    Xclean = X.copy()
    X += np.random.standard_normal(X.shape) * noise
    
    
    # Rotate to high-dimensional space
    Q = np.random.standard_normal((64,2))
    Q,_ = np.linalg.qr(Q)
    
    X = X.T.dot(Q.T) # rotate
    Xclean = Xclean.T.dot(Q.T)
    
    # scale 
    X = 2 * (X - np.min(X)) / np.ptp(X) - 1
    Xclean = 2 * (Xclean - np.min(Xclean)) / np.ptp(Xclean) - 1

    
    # split into train and test set 
    X_train = X[0:600]   
    X_test = X[600:]

    X_train_clean = Xclean[0:600]   
    X_test_clean = Xclean[600:]     
    
    #******************************************************************************
    # Return train and test set
    #******************************************************************************
    return X_train, X_test, X_train_clean, X_test_clean, 64, 1

def pendulum_timeseries(noise, theta=2.4):
    
    np.random.seed(1)

    def sol(t,theta0):
        S = np.sin(0.5*(theta0) )
        K_S = ellipk(S**2)
        omega_0 = np.sqrt(9.81)
        sn,cn,dn,ph = ellipj( K_S - omega_0*t, S**2 )
        theta = 2.0*np.arcsin( S*sn )
        d_sn_du = cn*dn
        d_sn_dt = -omega_0 * d_sn_du
        d_theta_dt = 2.0*S*d_sn_dt / np.sqrt(1.0-(S*sn)**2)
        return np.stack([theta, d_theta_dt],axis=1)
    
    
    anal_ts = np.arange(0, 2200*0.1, 0.1)
    X = sol(anal_ts, theta)
    
    X = X.T
    Xclean = X.copy()
    X += np.random.standard_normal(X.shape) * noise
    
    
    # Rotate to high-dimensional space
    # Q = np.random.standard_normal((64,2))
    # Q,_ = np.linalg.qr(Q)
    
    # X = X.T.dot(Q.T) # rotate
    # Xclean = Xclean.T.dot(Q.T)
    
    # scale 
    X = 2 * (X - np.min(X)) / np.ptp(X) - 1
    Xclean = 2 * (Xclean - np.min(Xclean)) / np.ptp(Xclean) - 1

    
    # split into train and test set 
    X_train = X[:,0:600].T   
    X_test = X[:,600:].T

    X_train_clean = Xclean[:,0:600].T
    X_test_clean = Xclean[:,600:].T 
    
    #******************************************************************************
    # Return train and test set
    #******************************************************************************
    return X_train, X_test, X_train_clean, X_test_clean, 2, 1

def flow(num):
    cylinder = np.load('/Users/yueenjiang/Desktop/Research/ST-Forecast/flow_cylinder.npy')

    Nt, m,n = cylinder.shape
    # T*x*y
    print(cylinder.shape)
    Xtrain = cylinder[:120, :200, :]
    Xtest = cylinder[120:, :200, :]

    Xtrain, Xtest = rescale(Xtrain, Xtest)

    X_train_clean = Xtrain.copy()
    X_test_clean = Xtest.copy()

    return Xtrain, Xtest, X_train_clean, X_test_clean, m, n

def flow_cylinder():
    X = np.load('/Users/yueenjiang/Desktop/Research/ST-Forecast/flow_cylinder.npy')
    print(X.shape)
   
    # Split into train and test set
    X = X[:, 65:, :]
    X = X[:, ::2, ::1]
   
    #X = X[:, ::6, ::3]
    #X = X[:, 0:64, 0:64]

    t, m, n = X.shape
 
    # mean subtract
    X = X.reshape(-1,m*n)
    X -= X.mean(axis=0)        
   
    # scale
    X = X.reshape(-1,m*n)
    X = 2 * (X - np.min(X)) / np.ptp(X) - 1
    X = X.reshape(-1,m,n)    
   
    # split into train and test set
   
    X_train = X[::120]  
    X_test = X[120::]
   
    #******************************************************************************
    # Return train and test set
    #******************************************************************************
    return X_train, X_test, X_train, X_test, m, n

def flow_cylinder_nosub():
    X = np.load('/Users/yueenjiang/Desktop/Research/ST-Forecast/flow_cylinder.npy')
    print(X.shape)
   
    # Split into train and test set
    # X = X[:, 65:, :]
    # X = X[:, ::2, ::1]

    t, m, n = X.shape
 
    # mean subtract
    # X = X.reshape(-1,m*n)
    # X -= X.mean(axis=0)
   
    # scale
    X = X.reshape(-1,m*n)
    X = 2 * (X - np.min(X)) / np.ptp(X) - 1
    X = X.reshape(-1,m,n)
   
    # split into train and test set
   
    X_train = X[:120, ]  
    X_test = X[120:, :]
   
    #******************************************************************************
    # Return train and test set
    #******************************************************************************
    return X_train, X_test, X_train, X_test, m, n

def h5data_load(fpath):
    hf = h5py.File(fpath, 'r')
    time = np.array(hf['time'])
    ss = np.array(hf['mu'])
    return time, ss

def avg_downsample(arr, n = 100):
    window = (1.0 / n) * np.ones(n,)
    res = np.convolve(arr, window, mode='valid')[::n]
    return res

def shear_stress(p):
    # fpath = '/Users/yueenjiang/Desktop/Research/Earthquake Forecast/Labquake/2022GRL-pData/p4677_shear_data_2MPa.hdf5'
    fpath = '/Users/yueenjiang/Desktop/Research/Earthquake Forecast/Labquake/2022-EPSL-AR/forecasting/' + p +'-newDF.npy'
    time, mu = h5data_load(fpath)

    dt = time[1]-time[0]
    time_new = np.arange(time[0], time[-1], dt*100)

    mu_new = np.interp(time_new, time, mu)

    mu_smooth = avg_downsample(mu, 100)
    X = mu_smooth

    split = int(max(mu_new.shape)*0.7)
    X_train = X[:split]
    X_test = X[split:]

    return X_train, X_test, X_train, X_test


def train_val_test_split(df, train_percentage, val_percentage, test_percentage,window_shift, batch_size,steps_in=200, steps_out=100):
  """Splits the input DataFrame in train, validation and test.

  Args:
  ----------
  df : DataFrame
        Input DataFrame
  train_percentage : float 0<=train_percentage<=1 (suggestion=0.7)
        Percentage of df length to use as training data.
  val_percentage : float 0<=val_percentage<=1 (suggestion=0.1)
        Percentage of df length to use as validation data.
  test_percentage : float 0<=test_percentage<=1 (suggestion=0.2)
        Percentage of df length to use as testing data.
        Note: train_percentage+val_percentage+test_percentage must be <1 and should be =1.
  window_shift : int (suggestion=10)
        Shift between one window and the following.
  batch_size : int
        Size of the batch used during the training of the model.
  steps_in : int (default=200)
        Number of points in input for the model, representing the known past.
  steps_out : int (default=100)
        Number of points in output from the model, representing the future to be forecasted.

  Returns:
  ----------
  tr_dl : torch.utils.data.DataLoader
        Resulting training set
  val_dl : torch.utils.data.DataLoader
        Resulting validation set
  test_dl : torch.utils.data.DataLoader
        Resulting testing set

  """
  train_size = int(len(df) * train_percentage)
  dataset_train = df[0:train_size]
  training_set = dataset_train.iloc[:, 0:dataset_train.shape[1]].values

  val_size = int(len(df) * val_percentage)
  dataset_val = df[train_size-steps_out:train_size+val_size+steps_out]
  val_set= dataset_val.iloc[:, 0:dataset_val.shape[1]].values

  test_size = int(len(df) * test_percentage)
  dataset_test = df[train_size+val_size-steps_out:train_size+val_size+test_size]
  test_set= dataset_test.iloc[:, 0:dataset_test.shape[1]].values

  print('training_set.shape: ',training_set.shape,' val_set.shape: ', val_set.shape,' test_set.shape: ', test_set.shape)

  df_tr = []
  for j in range(0,training_set.shape[1]):
    for i in range(0,training_set.shape[0] - steps_in - steps_out , window_shift):
      df_tr.append(training_set[:,j][i:i+steps_in+steps_out])

  X_train = []

  for elem in df_tr:
      X_train.append(elem[0:steps_in + steps_out])
  X_train = np.array(X_train)
  print("X_train.shape: ", X_train.shape)

  df_val = []
  for j in range(0,val_set.shape[1]):
    for i in range(0,val_set.shape[0] - steps_in - steps_out , window_shift):
      df_val.append(val_set[:,j][i:i+steps_in+steps_out])

  X_val = []
  for elem in df_val:
      X_val.append(elem[0:steps_in + steps_out])
  X_val = np.array(X_val)
  print("X_val.shape: ", X_val.shape)

#   df_test = []
#   for j in range(0,test_set.shape[1]):
#     for i in range(0,test_set.shape[0] - steps_in - steps_out , window_shift):
#       df_test.append(test_set[:,j][i:i+steps_in+steps_out])

#   X_test = []
#   for elem in df_test:
#       X_test.append(elem[0:steps_in + steps_out])
#   X_test = np.array(X_test)
#   print("X_test.shape: ", X_test.shape)

  return X_train, X_val, test_set
#   src_train= torch.from_numpy(X_train)
#   src_val = torch.from_numpy(X_val)
#   src_test = torch.from_numpy(X_test)

#   train_dataset = TensorDataset(src_train)
#   val_dataset = TensorDataset(src_val)
#   test_dataset = TensorDataset(src_test)

#   tr_dl = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size, shuffle=True, num_workers=0, drop_last=True)
#   val_dl = torch.utils.data.DataLoader(val_dataset, batch_size=batch_size, shuffle=True, num_workers=0, drop_last=True)
#   test_dl = torch.utils.data.DataLoader(test_dataset, batch_size=batch_size, shuffle=False, num_workers=0, drop_last=True)

#   return tr_dl, val_dl, test_dl